#define GTK_ENABLE_BROKEN

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "main.h"

#include "io.h"

/* undo functionality */

int find_rr(stl_string dom, stl_string ttl, stl_string type, stl_string data) {
  GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
  int ix = 0;
  char *_dom, *_ttl, *_type, *_data;
  while (true) {
    int ret = gtk_clist_get_text(list, ix, 0, &_dom);
    if (ret != 1)
      return -1;
    gtk_clist_get_text(list, ix, 1, &_ttl);
    gtk_clist_get_text(list, ix, 2, &_type);
    gtk_clist_get_text(list, ix, 3, &_data);
    if (dom == _dom && ttl == _ttl && type == _type && data == _data)
      return ix;
    ix++;
  }
}

class Action {
 public:
  virtual ~Action() { };
  virtual void revert() { };
  virtual stl_string getdesc() { return ""; };
  virtual Action *redo_action() { return new Action(); };
};

class DelAction : public Action {
 public:
  stl_string dom, ttl, type, data;
  stl_string op;
  DelAction(int _ix, stl_string _op = "Remove");
  void revert();
  stl_string getdesc();
  Action *redo_action();
};

class AddAction : public Action {
 public:
  int ix;
  stl_string nm,tt,ty,dt;
  stl_string op;
  AddAction(stl_string s1, stl_string s2, stl_string s3, stl_string s4, stl_string _op = "Add") {
    nm=s1;tt=s2;ty=s3;dt=s4;
    op = _op; ix = -1;
  }
  AddAction(int _ix, stl_string _op = "Add") { op = _op; ix = _ix; }
  void revert() {
    if (ix == -1) ix = find_rr(nm,tt,ty,dt);
    if (ix != -1) {
      gtk_clist_remove(GTK_CLIST(lookup_widget(frmMain, "lstRrs")), ix);
      set_edited();
    }
  }
  stl_string getdesc() {
    char *type;
    if (ix == -1) ix = find_rr(nm,tt,ty,dt);
    if (ix != -1) {
      gtk_clist_get_text(GTK_CLIST(lookup_widget(frmMain, "lstRrs")), ix, 2, &type);
    } else type = "unknown";
    return op + " " + type + " record";
  }
  Action *redo_action() {
    if (op == "Add") return new DelAction(ix, "Add"); else return new DelAction(ix, "Remove");
  }
};

DelAction::DelAction(int _ix, stl_string _op) {
    GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
    char *tmp;
    op = _op;
    gtk_clist_get_text(list, _ix, 0, &tmp); dom = tmp;
    gtk_clist_get_text(list, _ix, 1, &tmp); ttl = tmp;
    gtk_clist_get_text(list, _ix, 2, &tmp); type = tmp;
    gtk_clist_get_text(list, _ix, 3, &tmp); data = tmp;    
  }
  void DelAction::revert() {
    GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
    const char *entries[4] = { dom.c_str(), ttl.c_str(), type.c_str(), data.c_str() };
    gtk_clist_append(list, (gchar**)entries);
    gtk_clist_sort(list);
    set_edited();
  }
stl_string DelAction::getdesc() {
    return op + " " + type + " record";
  }
Action *DelAction::redo_action() {
    if (op == "Add")
      return new AddAction(dom, ttl, type, data, "Add");
    else
      return new AddAction(dom, ttl, type, data, "Remove");
  }

class EditAction : public Action {
 public:
  stl_string dom, ttl, type, data;
  int ix;
  EditAction(int _ix) {
    GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
    char *tmp;
    ix = _ix;
    gtk_clist_get_text(list, _ix, 0, &tmp); dom = tmp;
    gtk_clist_get_text(list, _ix, 1, &tmp); ttl = tmp;
    gtk_clist_get_text(list, _ix, 2, &tmp); type = tmp;
    gtk_clist_get_text(list, _ix, 3, &tmp); data = tmp;
  }
  void revert() {
    GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
    gtk_clist_set_text(list, ix, 0, dom.c_str());
    gtk_clist_set_text(list, ix, 1, ttl.c_str());
    gtk_clist_set_text(list, ix, 1, type.c_str());
    gtk_clist_set_text(list, ix, 3, data.c_str());
    gtk_clist_sort(list);
    set_edited();
  }
  stl_string getdesc() {
    return stl_string("Edit ") + type + " RR";
  }
  Action *redo_action() {
    return new EditAction(ix);
  }
};

class ZcAction : public Action {
 public:
  stl_string zconfig;
  ZcAction() {
    zconfig = zone_data;
  }
  void revert() {
    zone_data = zconfig;
    if (frmZoneConfig) {
      gtk_editable_delete_text(GTK_EDITABLE(lookup_widget(frmZoneConfig, "txtConfig")), 0, -1);
      gtk_text_insert(GTK_TEXT(lookup_widget(frmZoneConfig, "txtConfig")),
                      NULL, NULL, NULL, zone_data.c_str(), -1);
    }
  }
  stl_string getdesc() {
    return "Edit zone configuration";
  }
  Action *redo_action() {
    return new ZcAction();
  }
};

stl_list(Action*) undo_list, redo_list;

void undolabels() {
  Action *a;
  GtkWidget *w;
  w = lookup_widget(frmMain, "mnuRRundo");
  if (!undo_list.empty()) {
    a = *undo_list.begin();
    gtk_widget_set_sensitive(GTK_WIDGET(w), TRUE);
    gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child),
      (stl_string("Undo last operation: ") + a->getdesc()).c_str());
  } else {    
    gtk_widget_set_sensitive(GTK_WIDGET(w), FALSE);
    gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), "Undo last operation");
  }
  w = lookup_widget(frmMain, "mnuRRredo");
  if (!redo_list.empty()) {
    a = *redo_list.begin();
    gtk_widget_set_sensitive(GTK_WIDGET(w), TRUE);
    gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child),
      (stl_string("Redo last operation: ") + a->getdesc()).c_str());
  } else {
    gtk_widget_set_sensitive(GTK_WIDGET(w), FALSE);
    gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), "Redo last operation");
  }
}

void undo_clear() {
  undo_list.clear();
  redo_list.clear();
  undolabels();
}  

void add_undo(Action *action) {
  redo_list.clear();
  undo_list.push_front(action);
  undolabels();
}

/* globals */

gint lstrrs_selecteditem = -1;
domainname znroot;
char filename[2048] = "";
bool do_quit = false;
bool do_new = false;
bool do_open = false;
enum { quit, newmf, openmf, openrecent } save_action;
GtkMenuItem *recentitem;
GtkWidget *frmAsksave = NULL;
bool save_succeeded = false;
bool is_empty_new = false;

bool do_update_serial = true;

void unset_edited() {
  char *caption = (char *)malloc(strlen(filename) + strlen("Master File Editor - ") + 1);
  strcpy(caption,"Master File Editor - ");
  strcat(caption, filename);
  gtk_window_set_title(GTK_WINDOW(frmMain), caption);
  free(caption);
  status = loaded;
}

void set_edited() {
  if (status == none) return;
  char *caption = (char *)malloc(strlen(filename) + strlen("Master File Editor - ") + strlen(" [*]") + 1);
  strcpy(caption,"Master File Editor - ");
  strcat(caption, filename);
  strcat(caption, " [*]");
  gtk_window_set_title(GTK_WINDOW(frmMain), caption);
  free(caption);
  status = edited;
}

/* ----- error message ----------------------------------------------------- */
void
on_cmdErrorOK_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_main_quit();
}


gboolean
on_frmError_delete_event               (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_main_quit();
  return FALSE;
}

void show_error_message(const char *error) {
  GtkWidget *form = create_frmError();
  gtk_label_set(GTK_LABEL(lookup_widget(form, "lblError")), error);
  gtk_widget_show(form);
  gtk_main();
  gtk_widget_destroy(form);
}

/* ----- 'new' dialog ------------------------------------------------------ */

GtkWidget *frmNew = NULL;

void
on_mnuFileNew_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == edited && !do_new) {
    do_new = false;
    save_action = newmf;
    frmAsksave = create_frmAsksave();
    gtk_widget_show(frmAsksave);
  } else {
    do_new = false;
    is_empty_new = false;
    dom_readonly = false;
    frmNew = create_frmNew();
    gtk_entry_set_editable(GTK_ENTRY(lookup_widget(frmNew, "txtRoot")), TRUE);
    gtk_widget_show(frmNew);
  }
}

void
on_file_select                         (GtkButton       *button,
                                        GtkFileSelection*sel)
{
  char *dir = (char*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(sel));
  dom_readonly = false;
  if (load_master_file(dir))
    gtk_widget_destroy(GTK_WIDGET(sel));
}

void
on_file_close                         (GtkButton       *button,
                                       GtkFileSelection*sel)
{
  gtk_widget_destroy(GTK_WIDGET(sel));
}

void
on_mnuFileOpen_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == edited && !do_open) {
    do_open = false;
    save_action = openmf;
    frmAsksave = create_frmAsksave();
    gtk_widget_show(frmAsksave);
  } else {
    GtkWidget *sel;
    do_open = false;
    dom_readonly = false;
    sel = gtk_file_selection_new("Open master file");
    if (pwd[0]) gtk_file_selection_set_filename(GTK_FILE_SELECTION(sel), pwd);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(sel)->ok_button), "clicked", (GtkSignalFunc)on_file_select, (gpointer)sel);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(sel)->cancel_button), "clicked", (GtkSignalFunc)on_file_close, (gpointer)sel);
    gtk_window_set_modal(GTK_WINDOW(sel), TRUE);
    gtk_widget_show(sel);
  }
}


void
on_mnuFileSave_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (!filename[0]) {
    on_mnuFileSaveas_activate(menuitem, user_data);
    return;
  } else {
    save_succeeded = save_master_file();
    if (save_succeeded) unset_edited();
  }
}

/* ----- save as dialog -------------------------------------------------------------- */
GtkWidget *frmOverwrite = NULL;
GtkWidget *saveas = NULL;

void
on_cmdOverwriteOK_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
  /* store directory */
  char *dir = (char*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(saveas));
  char *ptr = dir + strlen(dir) - 1;
  while (ptr >= dir && *ptr != PATHDELIMC) ptr--;
  if (ptr - dir + 1 < (signed)sizeof(pwd)) {
    strncpy(pwd, dir, ptr - dir + 1);
    pwd[ptr-dir+1] = '\0';
  }
  if (save_master_file(dir)) {
    unset_edited();
    gtk_widget_destroy(GTK_WIDGET(saveas));
    if (frmOverwrite) {
      gtk_widget_destroy(GTK_WIDGET(frmOverwrite));
      frmOverwrite = NULL;
    }
  } else if (frmOverwrite) {
    gtk_widget_destroy(frmOverwrite);
    frmOverwrite = NULL;
  }
}

void
on_cmdOverwriteCancel_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy(frmOverwrite);
  frmOverwrite = NULL;
}


void
on_saveas_select                         (GtkButton       *button,
                                          GtkFileSelection*saveas)
{
  char *file = (char*)gtk_file_selection_get_filename(saveas);
  if (file_exists(file)) {
    frmOverwrite = create_frmOverwrite();
    gtk_widget_show(frmOverwrite);
  } else {
    on_cmdOverwriteOK_clicked((GtkButton*)button, NULL);
  }
}

void
on_saveas_close                         (GtkButton       *button,
                                         GtkFileSelection*saveas)
{
  gtk_widget_destroy(GTK_WIDGET(saveas));
}

void
on_mnuFileSaveas_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (!syntax_check()) return;
  saveas = gtk_file_selection_new("Save master file");
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(saveas), pwd);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(saveas)->ok_button), "clicked", (GtkSignalFunc)on_saveas_select, (gpointer)saveas);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(saveas)->cancel_button), "clicked", (GtkSignalFunc)on_saveas_close, (gpointer)saveas);
  gtk_window_set_modal(GTK_WINDOW(saveas), TRUE);
  gtk_widget_show(saveas);
}


void
on_mnuFileExit_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  on_frmMain_delete_event((GtkWidget*)NULL, (GdkEvent*)NULL, (gpointer)NULL);
}

void
on_cmdFileRecent_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  char *file = NULL;
  if (status == edited && !do_open) {
    do_open = false;
    save_action = openrecent;
    recentitem = menuitem;
    frmAsksave = create_frmAsksave();
    gtk_widget_show(frmAsksave);
  } else {
    do_open = false;
    gtk_label_get(GTK_LABEL(GTK_BIN(menuitem)->child), &file);
    if (file) load_master_file(file);
  }
}



void
on_mnuHelpAbout_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

  if (!frmAbout) {
    frmAbout = create_frmAbout();
    gtk_widget_show(frmAbout);
  } else {
    gtk_widget_hide(frmAbout);
  }
  gtk_widget_show(frmAbout);
}


GtkWidget *editrr;
bool is_new;
uint32_t edittype;
char nttl[32] = "";

void
on_mnuRrEdit_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  char *type, *dom, *ttl, *data;
  stl_string entry;
  GtkCList *list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  rr_type *rrtype = NULL;
  char buff[256];
  if (gtk_clist_get_text(list, lstrrs_selecteditem, 2, &type) == 1) {
    gtk_clist_get_text(list, lstrrs_selecteditem, 0, &dom);
    gtk_clist_get_text(list, lstrrs_selecteditem, 1, &ttl);
    gtk_clist_get_text(list, lstrrs_selecteditem, 3, &data);
    rrtype = rrtype_getinfo(type);
    if (!rrtype) {
      show_error_message("Internal error: RR not supported!");
      return;
    }    
    if (strcmpi(type, "SOA") == 0) {
      editrr = create_frmRR_SOA(); is_new = false; edittype = DNS_TYPE_SOA;
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")), dom);
      gtk_entry_set_editable(GTK_ENTRY(lookup_widget(editrr, "txtDom")), FALSE);
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), ttl);
      try {
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtMname")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtRname")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtSerial")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtRefresh")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtRetry")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtExpire")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtMinimum")), entry.c_str());
      } catch (PException p) { }
      gtk_widget_show(editrr);
    } else if (rrtype->type == DNS_TYPE_SRV) {
      editrr = create_frmRR_SRV(); is_new = false; edittype = DNS_TYPE_SRV;
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")), dom);
      gtk_entry_set_editable(GTK_ENTRY(lookup_widget(editrr, "txtDom")), FALSE);
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), ttl);
      try {
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtPriority")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtWeight")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtPort")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtHost")), entry.c_str());
      } catch (PException p) { }
      gtk_widget_show(editrr);      
    } else if (rrtype->type == DNS_TYPE_MX || rrtype->type == DNS_TYPE_HINFO || rrtype->type == DNS_TYPE_RP) {
      if (rrtype->type == DNS_TYPE_MX) on_mnuRR_MX_activate((GtkMenuItem*)NULL, NULL);
      if (rrtype->type == DNS_TYPE_HINFO) on_mnuRR_HINFO_activate((GtkMenuItem*)NULL, NULL);
      if (rrtype->type == DNS_TYPE_RP) on_mnuRR_RP_activate((GtkMenuItem*)NULL, NULL);
      is_new = false; edittype = rrtype->type;
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")), dom);
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), ttl);
      try {
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtData1")), entry.c_str());
        entry = read_entry(data);
        gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtData2")), entry.c_str());
      } catch (PException p) { }        
      sprintf(buff, "Edit %s record: %s", rrtype->name, dom);
      gtk_window_set_title(GTK_WINDOW(editrr), buff);
      gtk_widget_show(editrr);
    } else {
      if (rrtype->type == DNS_TYPE_A) on_mnuRR_A_activate((GtkMenuItem*)NULL, NULL);
      else if (rrtype->type == DNS_TYPE_AAAA) on_mnuRR_AAAA_activate((GtkMenuItem*)NULL, NULL);
      else if (rrtype->type == DNS_TYPE_NS) on_mnuRR_NS_activate((GtkMenuItem*)NULL, NULL);
      else if (rrtype->type == DNS_TYPE_CNAME) on_mnuRR_CNAME_activate((GtkMenuItem*)NULL, NULL);
      else if (rrtype->type == DNS_TYPE_PTR) on_mnuRR_PTR_activate((GtkMenuItem*)NULL, NULL);
      else if (rrtype->type == DNS_TYPE_TXT) on_mnuRR_TXT_activate((GtkMenuItem*)NULL, NULL);
      else if (rrtype->type == DNS_TYPE_LOC) on_mnuRR_LOC_activate((GtkMenuItem*)NULL, NULL);
      else editrr = create_frmRR_Generic();
      is_new = false; edittype = rrtype->type;
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")), dom);
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), ttl);
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtData")), data);
      sprintf(buff, "Edit %s record: %s", rrtype->name, dom);
      gtk_window_set_title(GTK_WINDOW(editrr), buff);
      gtk_widget_show(editrr);
    }
  }
}


void
on_mnuRrRemove_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status != none) {
    char *type;
    if (gtk_clist_get_text(GTK_CLIST(lookup_widget(frmMain, "lstRrs")), lstrrs_selecteditem, 2, &type) != 1) return;
    if (strcmpi(type, "SOA") == 0  && !is_cachefile) {
      show_error_message("The SOA record is obligatory and can't be removed!");
    } else {
      add_undo(new DelAction(lstrrs_selecteditem));
      gtk_clist_remove(GTK_CLIST(lookup_widget(frmMain, "lstRrs")), lstrrs_selecteditem);
      set_edited();
    }
  }
}

void
on_cmdNew_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
  on_mnuFileNew_activate((GtkMenuItem*)NULL, NULL);
}


void
on_cmdOpen_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
  on_mnuFileOpen_activate((GtkMenuItem*)NULL, NULL);
}


void
on_cmdSave_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
  on_mnuFileSave_activate((GtkMenuItem*)NULL, NULL);
}


void
on_cmdInsert_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_menu_popup(GTK_MENU(lookup_widget(frmMain, "mnuRR_menu")), NULL, NULL, NULL, NULL, 0, 0);
}


void
on_cmdRemove_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  on_mnuRrRemove_activate((GtkMenuItem*)NULL, NULL);
}


gboolean
on_frmAbout_delete_event               (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_destroy(frmAbout);
  frmAbout = (GtkWidget *)NULL;
  return FALSE;
}

void
on_cmdClose_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy(frmAbout);
  frmAbout = (GtkWidget *)NULL;
}

void
on_cmdConfig_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  if (status != none) {
    if (!frmZoneConfig) {
      frmZoneConfig = create_frmZoneConfig();
      gtk_text_insert(GTK_TEXT(lookup_widget(frmZoneConfig, "txtConfig")),
                      NULL, NULL, NULL, zone_data.c_str(), -1);
      gtk_widget_show(frmZoneConfig);
    } else {
      gtk_widget_hide(frmZoneConfig);
    }
    gtk_widget_show(frmZoneConfig);
  }
}


gboolean
on_frmZoneconfig_delete_event          (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_destroy(frmZoneConfig);
  frmZoneConfig = NULL;
  return FALSE;
}


void
on_cmdOK_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
  gchar *data = gtk_editable_get_chars(GTK_EDITABLE(lookup_widget(frmZoneConfig, "txtConfig")), 0, -1);
  add_undo(new ZcAction());
  zone_data = data;
  set_edited();
  g_free(data);
  gtk_widget_destroy(frmZoneConfig);
  frmZoneConfig = NULL;
}


void
on_cmdCancel_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy(frmZoneConfig);
  frmZoneConfig = NULL;
}


void
on_cmdSyntax_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  syntax_check();
}

void
on_mnuRR_A_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_A;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "The A record maps a domain name to an IPv4 IP number. "
                     "Enter the IP number in dotted-quad form (x.x.x.x) here.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData")),
                     "IPv4 number");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData"),
                     "IP number; for example: 192.168.1.2", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add A record");
  gtk_widget_show(editrr);

}


void
on_mnuRR_AAAA_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_AAAA;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "The AAAA record maps a domain name to an IPv6 IP number. "
                     "IPv6 is the successor of IPv4; IPv6 addresses look like "
                     "1:dead::beef.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData")),
                     "IPv6 number");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData"),
                     "IPv6 number; for example: 1:dead::beef", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add AAAA record");
  gtk_widget_show(editrr);

}


void
on_mnuRR_NS_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_NS;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "The NS record lists the nameservers for a given zone. "
                     "You should have these at the top of your zone, and you "
                     "can use them to delegate authority as well.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData")),
                     "Nameserver domain name:");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData"),
                     "Domain name of the nameserver for the zone", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add NS record");
  gtk_widget_show(editrr);

}


void
on_mnuRR_CNAME_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_CNAME;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "Defines the domain name it is attached to as an alias "
                     "to another domain. If a query is received for such a "
                     "domain, it will be redirected to the specified target.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData")),
                     "Target domain:");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData"),
                     "Domain name of the alias target", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add CNAME record");
  gtk_widget_show(editrr);

}


void
on_mnuRR_MX_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Two(); is_new = true; edittype = DNS_TYPE_MX;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "Use the MX record to define mail servers for the domain "
                     "name. If mail is received for <user>@<thisdomain>, it "
                     "will be sent to the mail server specified here. You can "
                     "create multiple mail servers records; the ones with the "
                     "lowest priority value will be tried first.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData1")),
                     "Priority value:");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData2")),
                     "Mail server domain:");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData1"),
                     "Priority value; MX records with lowest priority value will be tried first. Only if none of one level can be reached, entries of a higher priority level will be tried.", NULL);
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData2"),
                     "Domain name of the mail server.", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add MX record");
  gtk_widget_show(editrr);

}


void
on_mnuRR_PTR_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_PTR;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "The PTR record is used to do reverse-mapping of IP "
                     "numbers to domain names in the in-addr.arpa. zone.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData")),
                     "Pointer target:");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData"),
                     "Domain name the IP number points to", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add PTR record");
  gtk_widget_show(editrr);

}


void
on_mnuRR_SRV_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_SRV(); is_new = true; edittype = DNS_TYPE_SRV;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add SRV record");
  gtk_widget_show(editrr);
}


void
on_mnuRR_TXT_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_TXT;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "The TXT record can be used to insert free-form text "
                     "into your zone. This information can be read using "
                     "DNS query tools such as posask and nslookup.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData")),
                     "Free-form text:");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData"),
                     "Basically, any kind of free-form text", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add TXT record");
  gtk_widget_show(editrr);

}


void
on_mnuRR_HINFO_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Two(); is_new = true; edittype = DNS_TYPE_HINFO;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "Using the HINFO RR, you can specify information about the "
                     "host type that is running on a particular host. Nowadays, "
                     "this RR type is used only very occasionally. If you use "
                     "spaces, be sure to use quotes around values, e.g. "
                     "\"foo bar\".");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData1")),
                     "Hardware type:");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData2")),
                     "Operating system:");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData1"),
                     "Hardware architecture of the system", NULL);
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData2"),
                     "Operating system running on the system", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add HINFO record");
  gtk_widget_show(editrr);
}


void
on_mnuRR_WKS_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_WKS;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add WKS record");
  gtk_widget_show(editrr);
}


void
on_mnuRR_RP_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Two(); is_new = true; edittype = DNS_TYPE_RP;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "Using the RP RR, you can specify a responsible person for "
                     "a domain name. This RR is not widely used. You can use TXT "
                     "record attached to the info domain to specify additional "
                     "information.");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData1")),
                     "Email address:");
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData2")),
                     "Info domain:");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData1"),
                     "Email address of the person responsible for this domain", NULL);
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData2"),
                     "Domain on which TXT records about the person can be found, or . if no domain is available.", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add RP record");
  gtk_widget_show(editrr);
}


void
on_mnuRR_AFSDB_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_AFSDB;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add AFSDB record");
  gtk_widget_show(editrr);
}


void
on_mnuRR_PX_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_PX;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add PX record");
  gtk_widget_show(editrr);
}


void
on_mnuRR_NAPTR_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_NAPTR;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add NAPTR record");
  gtk_widget_show(editrr);
}

void
on_mnuRR_LOC_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_Generic(); is_new = true; edittype = DNS_TYPE_LOC;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblHelp")),
                     "The LOC record specifies the physical location for a host."
                     "The location RR consists of the following fields:\n"
                     "  - Latitude in degrees, e.g. 42 21 54 N\n"
                     "  - Longitude in degrees, e.g. 71 06 18 W\n"
                     "  - Altitude to sea level in meters, e.g. -24m\n"
                     "  - Size in meteers, e.g. 1m (defaults to 1m)\n"
                     "  - Horizontal (def. 10000m) and vertical (def. 10m) precision\n"
                     "Some examples:\n"
                     "  - 42 21 54 N 71 06 18 W -24m 30m\n"
                     "  - 42 21 28.764 N 71 00 51.617 W -44m 2000m\n"
                     );

  gtk_label_set_text(GTK_LABEL(lookup_widget(editrr, "lblData")),
                     "Location information");
  GtkTooltips *tips = (GtkTooltips*)gtk_object_get_data(GTK_OBJECT(editrr), "tooltips");
  gtk_tooltips_set_tip (tips, lookup_widget(editrr, "txtData"),
                     "Location information", NULL);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add LOC record");
  gtk_widget_show(editrr);
}


void
on_mnuRR_SOA_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  editrr = create_frmRR_SOA(); is_new = true; edittype = DNS_TYPE_SOA;
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")), "@");
  gtk_entry_set_editable(GTK_ENTRY(lookup_widget(editrr, "txtDom")), FALSE);
  gtk_entry_set_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")), nttl);
  gtk_window_set_title(GTK_WINDOW(editrr), "Add SOA record");
  gtk_widget_show(editrr);
}




void
on_lstRrs_select_row                   (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  lstrrs_selecteditem = row;
}


gboolean
on_lstRrs_button_press_event           (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  if (event->type == GDK_2BUTTON_PRESS)
    on_mnuRrEdit_activate(NULL, (gpointer)NULL);
  return FALSE;
}



void
on_txtRoot_changed                     (GtkEditable     *editable,
                                        gpointer         user_data)
{
  char *text = (char*)gtk_entry_get_text(GTK_ENTRY(editable));
  set_edited();
  try {
    znroot = text;
  } catch (PException p) { }
}


void
on_mnuToolsRelative_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  set_edited();
  zone_makerelative(true);
}


void
on_mnuToolsAbsolute_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (status == none) return;
  set_edited();
  zone_makerelative(false);
}

gboolean
on_frmMain_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  if (do_quit || status != edited) { gtk_main_quit(); return TRUE; }
  save_action = quit;
  frmAsksave = create_frmAsksave();
  gtk_widget_show(frmAsksave);
  return TRUE;
}


void
on_cmdSaveOK_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  on_mnuFileSave_activate((GtkMenuItem*)NULL, NULL);
  gtk_widget_destroy(frmAsksave);
  if (save_action == quit) {
    if (save_succeeded) gtk_main_quit();
  } else if (save_action == openmf) {
    if (save_succeeded) {
      do_open = true;
      on_mnuFileOpen_activate((GtkMenuItem*)NULL, NULL);
    }
  } else if (save_action == newmf) {
    if (save_succeeded) {
      do_new = true;
      on_mnuFileNew_activate((GtkMenuItem*)NULL, NULL);
    }
  } else if (save_action == openrecent) {
    if (save_succeeded) on_cmdFileRecent_activate(recentitem, NULL);
  }
    
}


void
on_cmdSaveCancel_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  if (save_action == quit) {
    do_quit = true;
    gtk_main_quit();
  } else if (save_action == openmf) {
    do_open = true;
    on_mnuFileOpen_activate((GtkMenuItem*)NULL, NULL);
    gtk_widget_destroy(frmAsksave);
  } else if (save_action == newmf) {
    do_new = true;
    on_mnuFileNew_activate((GtkMenuItem*)NULL, NULL);
    gtk_widget_destroy(frmAsksave);
  } else if (save_action == openrecent) {
    do_open = true;
    on_cmdFileRecent_activate(recentitem, NULL);
    gtk_widget_destroy(frmAsksave);
  }
}


void
on_cmdNewOK_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
  char *ptr;
  char *entries[4];
  domainname tmpd;
  char buff[1024];

  try {
    if (!is_empty_new) discard_current_data();

    /* root domain */
    ptr = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmNew, "txtRoot")));
    if (ptr[0] == 0) throw PException("Please enter a zone root!");
    tmpd = domainname(ptr);
    gtk_entry_set_text(GTK_ENTRY(lookup_widget(frmMain, "txtRoot")), ptr);

    /* admin email */
    strcpy(nttl, "2h");
    ptr = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmNew, "txtAdmin")));
    entries[0] = "@"; entries[1] = "2h"; entries[2] = "SOA";
    if (ptr[0] != 0) {
      tmpd = domainname(ptr);
      sprintf(buff, "ns1 %s 1 2h 1h 2d 2h", ptr);
      entries[3] = buff;
    } else {
      entries[3] = DEFAULT_SOA_RECORD;
    }
    gtk_clist_append(list, (gchar**) entries);

    /* our ip number */
    ptr = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmNew, "txtDNS")));
    if (ptr[0] != 0) {
      txt_to_ip(buff, ptr);
      entries[0] = "@"; entries[1] = "2h"; entries[2] = "NS"; entries[3] = "ns1";
      gtk_clist_append(list, (gchar**) entries);
      entries[0] = "ns1"; entries[1] = "2h"; entries[2] = "A"; entries[3] = ptr;
      gtk_clist_append(list, (gchar**) entries);
    }

    /* secondary dns */
    ptr = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmNew, "txtSecondary")));
    if (ptr[0] != 0) {
      txt_to_ip(buff, ptr);
      entries[0] = "@"; entries[1] = "2h"; entries[2] = "NS"; entries[3] = "ns2";
      gtk_clist_append(list, (gchar**) entries);
      entries[0] = "ns2"; entries[1] = "2h"; entries[2] = "A"; entries[3] = ptr;
      gtk_clist_append(list, (gchar**) entries);
    }

    /* mail server */
    ptr = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmNew, "txtMail")));
    if (ptr[0] != 0) {
      txt_to_ip(buff, ptr);
      entries[0] = "@"; entries[1] = "2h"; entries[2] = "MX"; entries[3] = "10 mail";
      gtk_clist_append(list, (gchar**) entries);
      entries[0] = "mail"; entries[1] = "2h"; entries[2] = "A"; entries[3] = ptr;
      gtk_clist_append(list, (gchar**) entries);
    }

    /* http server */
    ptr = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmNew, "txtHttp")));
    if (ptr[0] != 0) {
      txt_to_ip(buff, ptr);
      entries[0] = "www"; entries[1] = "2h"; entries[2] = "A"; entries[3] = ptr;
      gtk_clist_append(list, (gchar**) entries);
    }

    /* ftp server */
    ptr = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmNew, "txtFtp")));
    if (ptr[0] != 0) {
      txt_to_ip(buff, ptr);
      entries[0] = "ftp"; entries[1] = "2h"; entries[2] = "A"; entries[3] = ptr;
      gtk_clist_append(list, (gchar**) entries);
    }

    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget(frmNew, "chkCatchall"))) == TRUE) {
      entries[0] = "*"; entries[1] = "2h"; entries[2] = "CNAME"; entries[3] = "www";
      gtk_clist_append(list, (gchar**) entries);
    }
  } catch (PException p) {
    show_error_message(p.message);
    discard_current_data();
    return;
  }
  status = edited;
  gtk_widget_destroy(frmNew);
}

void
on_cmdNewCancel_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  if (is_empty_new) discard_current_data();
  gtk_widget_destroy(frmNew);
}

void
on_cmdSoaOk_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  stl_string ret;
  char *ttl, *tmp;
  uint32_t vttl;
  domainname vdom;
  stl_string vdat;
  GtkCList *list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  try {
    ttl = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")));
    vdom = gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")));
    vttl = txt_to_int(ttl);
    ret = stl_string(gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtMname")))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtRname"))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtSerial"))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtRefresh"))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtRetry"))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtExpire"))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtMinimum")));
    tmp = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtMinimum")));
    if (strlen(tmp) < 31) strcpy(nttl, tmp);
    rr_fromstring(DNS_TYPE_SOA, (char *)ret.c_str());

    if (is_new) {
      const char *entries[4];
      entries[0] = "@";
      entries[1] = ttl;
      entries[2] = "SOA";
      entries[3] = ret.c_str();
      gtk_clist_append(list, (gchar**)entries);
      add_undo(new AddAction(find_rr(entries[0], entries[1], entries[2], entries[3])));
    } else {
      add_undo(new EditAction(lstrrs_selecteditem));
      gtk_clist_set_text(list, lstrrs_selecteditem, 1, ttl);
      gtk_clist_set_text(list, lstrrs_selecteditem, 3, ret.c_str());
    }
    gtk_widget_destroy(editrr);
    gtk_clist_sort(list);
    set_edited();
  } catch (PException p) {
    show_error_message(p.message);
  }
}


void
on_cmdSoaCancel_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy(editrr);
}


void
on_cmdGenericOk_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  char *ret;
  char *ttl, *dom;
  uint32_t vttl;
  domainname vdom;
  stl_string vdat;
  GtkCList *list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  try {
    ttl = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")));
    dom = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")));
    ret = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtData")));
    /* test validity */
    vdom = domainname(dom, znroot);
    vttl = txt_to_int(ttl);
    rr_fromstring(edittype, ret);
    stl_string tmpdom = vdom.torelstring(znroot);
    /* apply data */
    if (is_new) {
      char *entries[4];
      entries[0] = (char*)tmpdom.c_str();
      entries[1] = ttl;
      entries[2] = rrtype_getinfo(edittype)->name;
      entries[3] = ret;
      gtk_clist_append(list, (gchar**)entries);
      add_undo(new AddAction(find_rr(entries[0], entries[1], entries[2], entries[3])));
    } else {
      add_undo(new EditAction(lstrrs_selecteditem));
      gtk_clist_set_text(list, lstrrs_selecteditem, 0, (char*)tmpdom.c_str());
      gtk_clist_set_text(list, lstrrs_selecteditem, 1, ttl);
      gtk_clist_set_text(list, lstrrs_selecteditem, 3, ret);
    }
    gtk_widget_destroy(editrr);
    gtk_clist_sort(list);
    set_edited();
  } catch (PException p) {
    show_error_message(p.message);
  }
}

void
on_cmdGenericCancel_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy(editrr);
}

void
on_cmdRR_Two_OK_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  char *ret1, *ret2;
  char *ttl, *dom;
  uint32_t vttl;
  domainname vdom;
  stl_string vdat, str;
  GtkCList *list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  try {
    ttl = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")));
    dom = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")));
    ret1 = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtData1")));
    ret2 = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtData2")));
    if (!ret1[0] || !ret2[0]) throw PException("Please fill in all fields!");
    str = stl_string(ret1) + " " + ret2;
    /* test validity */
    vdom = domainname(dom, znroot);
    vttl = txt_to_int(ttl);
    rr_fromstring(edittype, (char*)str.c_str());
    stl_string tmpdom = vdom.torelstring(znroot);
    /* apply data */
    if (is_new) {
      char *entries[4];
      entries[0] = (char*)tmpdom.c_str();
      entries[1] = ttl;
      entries[2] = rrtype_getinfo(edittype)->name;
      entries[3] = (char*)str.c_str();
      gtk_clist_append(list, (gchar**)entries);
      add_undo(new AddAction(find_rr(entries[0], entries[1], entries[2], entries[3])));
    } else {
      add_undo(new EditAction(lstrrs_selecteditem));
      gtk_clist_set_text(list, lstrrs_selecteditem, 0, (char*)tmpdom.c_str());
      gtk_clist_set_text(list, lstrrs_selecteditem, 1, ttl);
      gtk_clist_set_text(list, lstrrs_selecteditem, 3, str.c_str());
    }
    gtk_widget_destroy(editrr);
    gtk_clist_sort(list);
    set_edited();
  } catch (PException p) {
    show_error_message(p.message);
  }
}


void
on_cmdRR_Two_Cancel_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy(editrr);
}


void
on_frmSRV_cmdOK_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  char *ttl, *dom;
  uint32_t vttl;
  domainname vdom;
  stl_string vdat, ret;
  GtkCList *list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  try {
    ttl = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtTTL")));
    dom = (char*)gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtDom")));
    /* test validity */
    vdom = domainname(dom, znroot);
    vttl = txt_to_int(ttl);
    ret = stl_string(gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtPriority")))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtWeight"))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtPort"))) + " " +
          gtk_entry_get_text(GTK_ENTRY(lookup_widget(editrr, "txtHost")));
    rr_fromstring(DNS_TYPE_SRV, (char *)ret.c_str());
    stl_string tmpdom = vdom.torelstring(znroot);
    /* apply data */
    if (is_new) {
      char *entries[4];
      entries[0] = (char*)tmpdom.c_str();
      entries[1] = ttl;
      entries[2] = rrtype_getinfo(edittype)->name;
      entries[3] = (char*)ret.c_str();
      gtk_clist_append(list, (gchar**)entries);
      add_undo(new AddAction(find_rr(entries[0], entries[1], entries[2], entries[3])));
    } else {
      add_undo(new EditAction(lstrrs_selecteditem));
      gtk_clist_set_text(list, lstrrs_selecteditem, 0, (char*)tmpdom.c_str());
      gtk_clist_set_text(list, lstrrs_selecteditem, 1, ttl);
      gtk_clist_set_text(list, lstrrs_selecteditem, 3, ret.c_str());
    }
    gtk_widget_destroy(editrr);
    gtk_clist_sort(list);
    set_edited();
  } catch (PException p) {
    show_error_message(p.message);
  }
}


void
on_frmSRV_cmdCancel_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy(editrr);
}

/* undo/redo */

void
on_mnuRRundo_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  Action *a = *undo_list.begin(), *b;
  if (!a) return;
  b = a->redo_action();
  a->revert();
  redo_list.push_front(b);
  undo_list.pop_front();
  delete a;
  undolabels();
  if (undo_list.empty()) unset_edited();
}


void
on_mnuRRredo_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  Action *a = *redo_list.begin(), *b;
  if (!a) return;
  b = a->redo_action();
  a->revert();
  undo_list.push_front(b);
  redo_list.pop_front();
  delete a;
  undolabels();
  set_edited();
}


void
on_mnuToolsAutoserial_activate         (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  do_update_serial = GTK_CHECK_MENU_ITEM(menuitem)->active;
}

void
on_mnuToolsCachefile_activate          (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  is_cachefile = GTK_CHECK_MENU_ITEM(menuitem)->active;
  set_edited();
}
