/***************************************************************************
                          io.cpp  -  i/o
                             -------------------
    begin                : vr feb 7 2003
    copyright            : (C) 2003 by meilof
    email                : meilof@athlon
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define GTK_ENABLE_BROKEN
#include <poslib/poslib.h>
#include <stdio.h>
#include <ctype.h>
#ifdef _WIN32
extern "C" { int getcwd(char*,int); };
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "callbacks.h"
#include "main.h"
#include "support.h"
#include "interface.h"
#include "io.h"

_status status = none;

stl_string zone_data;
uint32_t last_serial = 0;
bool is_cachefile = false;

void discard_current_data() {
  strcpy(filename, "");
  zone_data = "";
  strcpy(nttl, "");
  if (frmZoneConfig)
    gtk_editable_delete_text(GTK_EDITABLE(lookup_widget(frmZoneConfig, "txtConfig")), 0, -1);
  /* clear the clist */
  gtk_clist_clear(GTK_CLIST(lookup_widget(frmMain, "lstRrs")));
  gtk_window_set_title(GTK_WINDOW(frmMain), "Master File Editor");
  status = none;
  is_cachefile = false;
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(lookup_widget(frmMain, "mnuToolsCachefile")), FALSE);
  gtk_entry_set_editable(GTK_ENTRY(lookup_widget(frmMain, "txtRoot")), !dom_readonly);
  undo_clear();
}

void zone_makerelative(bool isrelative) {
  int rr_ix = 0;
  gint lret;
  char *dom, *type, *data;
  rr_type *rrtype;
  char *ptr, *cptr;
  GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
  stl_string tmp, ret;

  try {
    while (1) {
      lret = gtk_clist_get_text(list, rr_ix, 0, &dom);
      if (lret != 1) return;

      gtk_clist_get_text(list, rr_ix, 2, &type);
      gtk_clist_get_text(list, rr_ix, 3, &data);

      rrtype = rrtype_getinfo(type);
      if (!rrtype) throw PException("Unsupported RR! Correct this error first!");
      
      /* check whether we can skip this rr altogether */
      if (strchr(rrtype->properties, 'd') == NULL &&
          strchr(rrtype->properties, 'm') == NULL) goto next_rr;

      rr_fromstring(rrtype->type, data);
      ptr = data;
      cptr = rrtype->properties;
      ret = "";

      /* FIXME: This asserts that properties do not have spaces in them.
         while this assertion is actually true for all current RR types,
         it is certainly a bit hackish */
      while (*cptr) {
        tmp = read_entry(ptr);
        if (*cptr == 'd' || *cptr == 'm') {
          /* domain name */
          if (isrelative) {
            if (ret.size()) ret += " ";
            ret += domainname(tmp.c_str(), znroot).torelstring(znroot);
          } else {
            if (ret.size()) ret += " ";
            ret += domainname(tmp.c_str(), znroot).tocstr();
          }
        } else {
          if (ret.size()) ret += " ";
          ret += tmp;
        }
        cptr++;
      }
      gtk_clist_set_text(list, rr_ix, 3, ret.c_str());
next_rr:
      rr_ix++;
    }
  } catch (PException p) {
    gtk_clist_select_row(list, rr_ix, 0);
    gtk_clist_moveto(list, rr_ix, 0, 0.5, 0);
    show_error_message(p.message);
  }
}

bool mf_is_edited;

void errorcallback(void *user_dat, const char *file, int linenum, const char *message) {
  show_error_message(message);
  mf_is_edited = true;
}


void rrliteralcallback(void *user_dat, const char *dom, const char *ttl, const char *type,
                       const char *rrdata, domainname origin) {
  GtkCList *list = (GtkCList *)user_dat;
  const char *entries[4];
  char *tmpa;
  stl_string rrcopy, tmp;
  char *rrdata2 = (char*)rrdata;
  
  /* replace tabs and multiple spaces by one space */
  bool quote;
  while (rrdata2 && rrdata2[0]) {
    if (*rrdata2 == '"') quote = true; else quote = false;
    tmp = read_entry(rrdata2);
    if (quote)
      rrcopy += stl_string("\"") + tmp + "\" ";
    else
      rrcopy += tmp + " ";
  }

  if (strcmpi(type, "SOA") == 0) {
    char *ptr = (char *)rrdata;
    stl_string str;
    try {
      read_entry(ptr); read_entry(ptr); /* mname, rname */
      last_serial = txt_to_int(read_entry(ptr).c_str()); /* serial number */
      read_entry(ptr); read_entry(ptr); read_entry(ptr); /* refresh, retru. expire*/
      str = read_entry(ptr); /* nttl value */
      if (strlen(str.c_str()) <= 31) {
        strcpy(nttl, str.c_str());
      }      
    } catch (PException p) { }
  } else if (gtk_clist_get_text(list, 0, 0, &tmpa) != 1) {
    if (is_cachefile) {
      strncpy(nttl, ttl, sizeof(nttl));
    } else {
      gtk_clist_append(list, (gchar**)entries);
      show_error_message("First record was not a SOA record. Using a default SOA record instead.");
      const char* entries2[4];
      entries2[0] = "@";
      entries2[1] = "2h";
      entries2[2] = "SOA";
      entries2[3] = DEFAULT_SOA_RECORD;
      gtk_clist_append(list, (gchar**)entries2);
    }
  }
  
  domainname tmpdom = domainname(dom, origin);
  entries[0] = (char*)tmpdom.torelstring(origin).c_str();
  entries[1] = ttl;
  entries[2] = type;
  entries[3] = rrcopy.c_str();
  gtk_clist_append(list, (gchar**)entries);
}

void commentcallback(void *user_dat, const char *comment) {
  if (strncmpi(comment, "set ", 4) == 0) {
    zone_data += (comment + 4);
    zone_data += "\n";
  } else if (strcmpi(comment, "mfedit is_cachefile") == 0) {
    is_cachefile = true;
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(lookup_widget(frmMain, "mnuToolsCachefile")), TRUE);
  }
}

void try_load_master_file(const char *file) {
  domainname znroot;
  try {
    read_master_file(file, znroot, lookup_widget(frmMain, "lstRrs"),
                     errorcallback, NULL, rrliteralcallback, commentcallback,
                     POSLIB_MF_AUTOPROBE|POSLIB_MF_NOSOA);
    gtk_entry_set_text(GTK_ENTRY(lookup_widget(frmMain, "txtRoot")), znroot.tocstr());
    if (mf_is_edited)
      show_error_message("Your master file contained errors. Please run a syntax check to repair them.");
  } catch (PException p) {
    discard_current_data();
    throw p;
  }
}

void recnam_down(int start, int nitems) {
  for (int t = start + nitems; t > start; t--) {
    recent[t] = recent[t - 1];
    recentnam[t] = recentnam[t - 1];
  }
}

void rec_rewidget() {
  for (int x = 0; x < nrecent; x++) {
    gtk_widget_hide (recent[x]);
    gtk_container_remove(GTK_CONTAINER(lookup_widget(frmMain, "mnuFile_menu")), recent[x]);
    gtk_container_add (GTK_CONTAINER(lookup_widget(frmMain, "mnuFile_menu")), recent[x]);
    if (recentnam[x]) {
      gtk_widget_show (recent[x]);
    }
  }
}

void recentfile(const char *file) {
  /* add file to recent list */
  GtkWidget *wtmp; char *ctmp;
  int t;
  for (t = 0; t < nrecent; t++) {
    if (!recentnam[t] || strcmpi(recentnam[t], file) == 0) break;
  }
  if (t == nrecent) t = nrecent - 1;

  gtk_label_set_text(GTK_LABEL(GTK_BIN(recent[t])->child), file);
  if (recentnam[t]) free(recentnam[t]);
  recentnam[t] = strdup(file);

  wtmp = recent[t];
  ctmp = recentnam[t];
  recnam_down(0, t);
  recent[0] = wtmp;
  recentnam[0] = ctmp;
  rec_rewidget();

  /* update pwd */
  char *ptr = (char *)file + strlen(file) - 1;
  while (ptr >= file && *ptr != PATHDELIMC) ptr--;
  if (ptr - file + 1 < (signed)sizeof(pwd)) {
    strncpy(pwd, file, ptr - file + 1);
    pwd[ptr-file+1] = '\0';
  }
}

bool load_master_file(const char *_file) {
  try {
    discard_current_data();
    mf_is_edited = false;

    /* see if we have a relative path */
    if (strchr(_file, PATHDELIMC) == NULL) {
      /* append pwd */
      getcwd(filename, PATH_MAX);
      if (strlen(filename) + strlen(_file) + 2 > PATH_MAX) {
        throw PException("File name too long!");
      }
      if (filename[strlen(filename)-1] != PATHDELIMC) strcat(filename, PATHDELIM);
      strcat(filename, _file);
    } else {
      if (strlen(filename) >= sizeof(filename)) throw PException("File name too long");
      strcpy(filename, _file);
    }

    recentfile(filename);
    if (file_exists(filename)) try_load_master_file(filename);
    unset_edited();
    if (mf_is_edited) set_edited();
    if (frmZoneConfig) {
      gtk_text_insert(GTK_TEXT(lookup_widget(frmZoneConfig, "txtConfig")),
                      NULL, NULL, NULL, zone_data.c_str(), -1);
    }
    char *tmp;
    if (!is_cachefile && gtk_clist_get_text(GTK_CLIST(lookup_widget(frmMain, "lstRrs")), 0, 0, &tmp) != 1) {
      /* apparently, an empty master file! */
      domainname znroot = guess_zone_name(filename);
      do_new = false;
      is_empty_new = true;
      frmNew = create_frmNew();
      gtk_entry_set_text(GTK_ENTRY(lookup_widget(frmNew, "txtRoot")), znroot.tocstr());
      gtk_entry_set_editable(GTK_ENTRY(lookup_widget(frmNew, "txtRoot")), !dom_readonly);
      gtk_widget_show(frmMain);
      gtk_widget_show(frmNew);
    }
  } catch (PException p) {
    show_error_message(p.message);
    return false;
  }
  return true;
}

#define check_any 1
#define check_delegated 2

bool check_dom(domainname &origin, domainname &dom, int flags) {
  int rr_ix = 0;

  char *cdom, *type;
  GtkCList *list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  gint ret;

  domainname zdom;

  while (1) {
    ret = gtk_clist_get_text(list, rr_ix, 0, &cdom);
    if (ret != 1) break;

    gtk_clist_get_text(list, rr_ix, 2, &type);

    try {
      zdom = domainname(cdom, origin);
      if (zdom == dom) {
        if (flags&check_any || strcmpi(type, "A") == 0 || strcmpi(type, "AAAA") == 0) return true;
      }
      if (dom >= zdom && zdom != origin && strcmpi(type, "NS") == 0 && !(flags&check_delegated)) {
        return true; /* delegation */
      }
    } catch (PException p) { }
    rr_ix++;
  }
  return false;
}

bool test_dnames(domainname &origin, rr_type *type, const char *rrdata, int flags) {
  int t = 0;
  char *p = type->properties;
  domainname test;
  while (*p) {
    if (*p == 'd') {
      test = rr_getdomain((char*)rrdata, type->type, t);
      if (test >= origin && !check_dom(origin, test, flags)) return false;
    }
    p++; t++;
  }
  return true;
}

bool find_dname(GtkCList *list, domainname &dom, domainname &origin, int n) {
  int ix = 0;
  char *cdom;
  domainname ddom;

  while (1) {
    if (gtk_clist_get_text(list, ix, 0, &cdom) != 1) break;
    if (ix != n) {
      ddom = domainname(cdom, origin);
      if (ddom == dom) return true;
    }
    ix++;
  }
  return false;
}

bool syntax_check() {
  domainname origin;

  if (status == none) {
    show_error_message("No master file loaded!");
    return false;
  }

  try {
    char *txt = (char *)gtk_entry_get_text(GTK_ENTRY(lookup_widget(frmMain, "txtRoot")));
    if (!txt[0]) throw PException("");
    origin = txt;
    znroot = origin;
  } catch (PException p) {
    show_error_message("Incorrect root domain for the zone!");
    return false;
  }

  int rr_ix = 0;
  char *dom, *ttl, *type, *data;
  GtkCList *list = NULL;
  gint ret;
  list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  domainname zdom;
  rr_type *ztype;
  int flags;
  stl_string rrstr;

  while (1) {
    ret = gtk_clist_get_text(list, rr_ix, 0, &dom);
    if (ret != 1) break;

    gtk_clist_get_text(list, rr_ix, 1, &ttl);
    gtk_clist_get_text(list, rr_ix, 2, &type);
    gtk_clist_get_text(list, rr_ix, 3, &data);

    try {
      zdom = domainname(dom, origin);
      if (!(zdom >= origin)) throw PException("Domain name outside of the zone!");
      ztype = rrtype_getinfo(type);
      if (!ztype) throw PException("Unknown RR type!");
      if (ztype->type == DNS_TYPE_SOA && rr_ix) throw PException("Zone may not contain multiple SOA records!");
      else if (!rr_ix && ztype->type != DNS_TYPE_SOA && !is_cachefile) throw PException("The first record of your zone should be a SOA record.");
      rrstr = rr_fromstring(ztype->type, data, origin.c_str());
      flags = 0;
      if (ztype->type == DNS_TYPE_CNAME || ztype->type == DNS_TYPE_DNAME ||
          ztype->type == DNS_TYPE_RP) flags = check_any;
      else if (ztype->type == DNS_TYPE_NS) flags = check_delegated;
      if (!is_cachefile && !test_dnames(origin, ztype, rrstr.c_str(), flags))
        throw PException("Incorrect domain name in RR data!\n\n"
                         " - The domain name may not exist; or\n"
                         " - The domain name may not have A or AAAA records and these records are required\n\n"
                         "(note that domain names without a trailing dot are relative to the zone root,\n"
                         "so if you want to point to a domain outside the zone, do something like \"www.acdam.net.\")");
      if (ztype->type == DNS_TYPE_CNAME && find_dname(list, zdom, origin, rr_ix)) throw PException("CNAME domain may not have other RRs as well!");
    } catch (PException p) {
      gtk_clist_select_row(list, rr_ix, 0);
      gtk_clist_moveto(list, rr_ix, 0, 0.5, 0);
      show_error_message(p.message);
      return false;
    }

    rr_ix++;
  }
  if (!rr_ix && !is_cachefile) {
    show_error_message("Your zone is completely empty!");
    return false;
  }
  return true;
}

FILE *try_fopen_w(const char *file) {
  struct stat st;

  if (stat(file, &st) != 0) return fopen(file, "w");
  else if (!S_ISREG(st.st_mode)) return NULL;
  else return fopen(file, "w");
}

bool save_master_file(const char *file) {
  char *dom, *type, *ttl, *data; int rr_ix = 0;
  GtkCList *list = (GtkCList *)lookup_widget(frmMain, "lstRrs");
  FILE *fout = NULL;

  if (status == none) {
    show_error_message("No master file loaded!");
    return false;
  }
  if (!syntax_check()) return false;

  if (file) {
    if (strlen(file) <= sizeof(filename)) {
      strcpy(filename, file);
    } else {
      show_error_message("File name too long!");
      return false;
    }
    fout = try_fopen_w(file);
    if (fout) recentfile(file);
  } else if (filename[0]) {
    fout = try_fopen_w(filename);
    if (fout) recentfile(filename);
  } else {
    show_error_message("No file!");
    return false;
  }

  if (!fout) {
    show_error_message("Could not open file for writing!");
    return false;
  }

  /* if nessecary, increase serial */
  if (do_update_serial && !is_cachefile) {
    try {
      GtkCList *list = GTK_CLIST(lookup_widget(frmMain, "lstRrs"));
      char *rrdata;
      stl_string retsoa, tmp;
      uint32_t serial;
      gtk_clist_get_text(list, 0, 3, &rrdata);

      /* skip mname, rname */
      retsoa += read_entry(rrdata) + " ";
      retsoa += read_entry(rrdata) + " ";

      serial = txt_to_int(read_entry(rrdata).c_str());

      if (serial == last_serial) { /* not updated manually */
        time_t tsecs = time(NULL);
        struct tm *tstruct = localtime(&tsecs);
        char buff[16];
        uint32_t newserial =
          (tstruct->tm_year + 1900) * 1000000 +
          (tstruct->tm_mon + 1) * 10000 +
          tstruct->tm_mday * 100;
        if (serial >= newserial) newserial = serial + 1;
        sprintf(buff, "%10d", newserial);
        retsoa += buff;
        retsoa += " ";
        retsoa += rrdata;

        gtk_clist_set_text(list, 0, 3, retsoa.c_str());
        last_serial = newserial;
      }
    } catch (PException p) {
      /* ah well, whatever */
      fprintf(stderr, "*** Error checking serial: %s\n", p.message);
    }
  }

  /* write master file */
  fprintf(fout, ";\n"
                "; DNS master file generated by Posadis mfedit\n"
                ";\n\n"
                "$origin %s\n\n", znroot.tocstr());

  if (is_cachefile) fprintf(fout, ";mfedit is_cachefile\n");

  /* write comments */
  char *orig = strdup(zone_data.c_str()), *cptr = orig, *ptr2;
  while (*cptr) {
    if ((ptr2 = strchr(cptr, '\n')))
      *ptr2 = '\0';
    else if ((ptr2 = strchr(cptr, '\r')))
      *ptr2 = '\0';
    fprintf(fout, ";set %s\n", cptr);
    if (!ptr2) break;
    ptr2++; while (*ptr2 && (*ptr2 == '\n' || *ptr2 == '\r')) ptr2++;
    cptr = ptr2;
  }
  free(orig);

  /* write rrs */
  while (1) {
    if (gtk_clist_get_text(list, rr_ix, 0, &dom) != 1) break;

    gtk_clist_get_text(list, rr_ix, 1, &ttl);
    gtk_clist_get_text(list, rr_ix, 2, &type);
    gtk_clist_get_text(list, rr_ix, 3, &data);

    fprintf(fout, "%-31s %-7s %-7s %s\n", dom, ttl, type, data);
    rr_ix++;
  }

  fclose(fout);

  return true;
}
